@@ -4,10 +4,9 @@
 #
 # Copyright (C) 2006-2007, 2009 Lukáš Lalinský
 # Copyright (C) 2009 Nikolai Prokoschenko
-# Copyright (C) 2009-2010, 2019 Philipp Wolfer
+# Copyright (C) 2009-2010, 2019-2020 Philipp Wolfer
 # Copyright (C) 2011-2013 Michael Wiencek
-# Copyright (C) 2013-2015, 2017-2019 Laurent Monin
-# Copyright (C) 2014 m42i
+# Copyright (C) 2013-2015, 2017-2020 Laurent Monin
 # Copyright (C) 2016-2017 Sambhav Kothari
 # Copyright (C) 2016-2017 Suhas
 # Copyright (C) 2018 Vishal Choudhary
@@ -27,23 +26,31 @@
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
 
-from PyQt5 import (
-    QtCore,
-    QtGui,
-)
+from PyQt5 import QtCore
 
 from picard import config
 from picard.const import PICARD_URLS
-from picard.script import ScriptParser
+from picard.script import (
+    ScriptParser,
+    script_function_documentation_all,
+)
 from picard.util import restore_method
 
+from picard.ui import (
+    FONT_FAMILY_MONOSPACE,
+    PicardDialog,
+)
 from picard.ui.moveable_list_view import MoveableListView
 from picard.ui.options import (
     OptionsCheckError,
     OptionsPage,
     register_options_page,
 )
+from picard.ui.theme import theme
 from picard.ui.ui_options_script import Ui_ScriptingOptionsPage
+from picard.ui.ui_scripting_documentation_dialog import (
+    Ui_ScriptingDocumentationDialog,
+)
 from picard.ui.widgets.scriptlistwidget import ScriptListWidgetItem
 
 
@@ -51,73 +58,80 @@ class ScriptCheckError(OptionsCheckError):
     pass
 
 
-class TaggerScriptSyntaxHighlighter(QtGui.QSyntaxHighlighter):
+DOCUMENTATION_HTML_TEMPLATE = '''
+<!DOCTYPE html>
+<html>
+<head>
+<style>
+dt {
+    color: %(script_function_fg)s
+}
+dd {
+    padding: 50px;
+    margin-bottom: 50px;
+}
+code {
+    font-family: %(monospace_font)s;
+}
+</style>
+</head>
+<body>
+    %(html)s
+</body>
+</html>
+'''
 
-    def __init__(self, document):
-        super().__init__(document)
-        self.func_re = QtCore.QRegExp(r"\$(?!noop)[a-zA-Z][_a-zA-Z0-9]*\(")
-        self.func_fmt = QtGui.QTextCharFormat()
-        self.func_fmt.setFontWeight(QtGui.QFont.Bold)
-        self.func_fmt.setForeground(QtCore.Qt.blue)
-        self.var_re = QtCore.QRegExp(r"%[_a-zA-Z0-9:]*%")
-        self.var_fmt = QtGui.QTextCharFormat()
-        self.var_fmt.setForeground(QtCore.Qt.darkCyan)
-        self.escape_re = QtCore.QRegExp(r"\\.")
-        self.escape_fmt = QtGui.QTextCharFormat()
-        self.escape_fmt.setForeground(QtCore.Qt.darkRed)
-        self.special_re = QtCore.QRegExp(r"[^\\][(),]")
-        self.special_fmt = QtGui.QTextCharFormat()
-        self.special_fmt.setForeground(QtCore.Qt.blue)
-        self.bracket_re = QtCore.QRegExp(r"[()]")
-        self.noop_re = QtCore.QRegExp(r"\$noop\(")
-        self.noop_fmt = QtGui.QTextCharFormat()
-        self.noop_fmt.setFontWeight(QtGui.QFont.Bold)
-        self.noop_fmt.setFontItalic(True)
-        self.noop_fmt.setForeground(QtCore.Qt.darkGray)
-        self.rules = [
-            (self.func_re, self.func_fmt, 0, -1),
-            (self.var_re, self.var_fmt, 0, 0),
-            (self.escape_re, self.escape_fmt, 0, 0),
-            (self.special_re, self.special_fmt, 1, -1),
-        ]
 
-    def highlightBlock(self, text):
-        self.setCurrentBlockState(0)
+class ScriptingDocumentationDialog(PicardDialog):
+    defaultsize = QtCore.QSize(570, 400)
 
-        for expr, fmt, a, b in self.rules:
-            index = expr.indexIn(text)
-            while index >= 0:
-                length = expr.matchedLength()
-                self.setFormat(index + a, length + b, fmt)
-                index = expr.indexIn(text, index + length + b)
+    def __init__(self, parent):
+        super().__init__(parent)
+        self.setWindowFlags(QtCore.Qt.Window)
+        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
+        self.parent = parent
+        self.parent.scripting_documentation_shown = True
+        self.setWindowFlags(QtCore.Qt.Window)
+        self.ui = Ui_ScriptingDocumentationDialog()
+        self.ui.setupUi(self)
+        args = {
+            "picard-doc-scripting-url": PICARD_URLS['doc_scripting'],
+        }
+        text = _('<a href="%(picard-doc-scripting-url)s">Open Scripting'
+                 ' Documentation in your browser</a>') % args
+        self.ui.scripting_doc_link.setText(text)
 
-        # Ignore everything if we're already in a noop function
-        index = self.noop_re.indexIn(text) if self.previousBlockState() <= 0 else 0
-        open_brackets = self.previousBlockState() if self.previousBlockState() > 0 else 0
-        while index >= 0:
-            next_index = self.bracket_re.indexIn(text, index)
+        def process_html(html, function):
+            if not html:
+                html = ''
+            template = '<dt>%s%s</dt><dd>%s</dd>'
+            if function.module is not None and function.module != 'picard.script':
+                module = ' [' + function.module + ']'
+            else:
+                module = ''
+            try:
+                firstline, remaining = html.split("\n", 1)
+                return template % (firstline, module, remaining)
+            except ValueError:
+                return template % ("<code>$%s()</code>" % function.name, module, html)
 
-            # Skip escaped brackets
-            if (next_index > 0) and text[next_index - 1] == '\\':
-                next_index += 1
+        funcdoc = script_function_documentation_all(
+            fmt='html',
+            postprocessor=process_html,
+        )
 
-            if (next_index > -1) and text[next_index] == '(':
-                open_brackets += 1
-            elif (next_index > -1) and text[next_index] == ')':
-                open_brackets -= 1
+        syntax_theme = theme.get_syntax_theme()
+        html = DOCUMENTATION_HTML_TEMPLATE % {
+            'html': "<dl>%s</dl>" % funcdoc,
+            'script_function_fg': syntax_theme.func.name(),
+            'monospace_font': FONT_FAMILY_MONOSPACE,
+        }
+        self.ui.textBrowser.setHtml(html)
+        self.ui.buttonBox.rejected.connect(self.close)
 
-            if next_index > -1:
-                self.setFormat(index, next_index - index + 1, self.noop_fmt)
-            elif (next_index == -1) and (open_brackets > 0):
-                self.setFormat(index, len(text) - index, self.noop_fmt)
-
-            # Check for next noop operation, necessary for multiple noops in one line
-            if open_brackets == 0:
-                next_index = self.noop_re.indexIn(text, next_index)
-
-            index = next_index + 1 if (next_index > -1) and (next_index < len(text)) else -1
-
-        self.setCurrentBlockState(open_brackets)
+    def closeEvent(self, event):
+        self.parent.scripting_documentation_shown = False
+        super().closeEvent(event)
 
 
 class ScriptingOptionsPage(OptionsPage):
@@ -139,15 +153,25 @@ class ScriptingOptionsPage(OptionsPage):
         super().__init__(parent)
         self.ui = Ui_ScriptingOptionsPage()
         self.ui.setupUi(self)
-        self.highlighter = TaggerScriptSyntaxHighlighter(self.ui.tagger_script.document())
         self.ui.tagger_script.setEnabled(False)
         self.ui.splitter.setStretchFactor(0, 1)
         self.ui.splitter.setStretchFactor(1, 2)
-        font = QtGui.QFont('Monospace')
-        font.setStyleHint(QtGui.QFont.TypeWriter)
-        self.ui.tagger_script.setFont(font)
         self.move_view = MoveableListView(self.ui.script_list, self.ui.move_up_button,
                                           self.ui.move_down_button)
+        self.ui.scripting_documentation_button.clicked.connect(self.show_scripting_documentation)
+        self.scripting_documentation_shown = None
+
+    def show_scripting_documentation(self):
+        if not self.scripting_documentation_shown:
+            self.scriptdoc_dialog = ScriptingDocumentationDialog(parent=self)
+            self.scriptdoc_dialog.show()
+        else:
+            self.scriptdoc_dialog.raise_()
+            self.scriptdoc_dialog.activateWindow()
+
+    def enable_tagger_scripts_toggled(self, on):
+        if on and self.ui.script_list.count() == 0:
+            self.ui.script_list.add_script()
 
     def script_selected(self):
         items = self.ui.script_list.selectedItems()
@@ -155,6 +179,7 @@ class ScriptingOptionsPage(OptionsPage):
             item = items[0]
             self.ui.tagger_script.setEnabled(True)
             self.ui.tagger_script.setText(item.script)
+            self.ui.tagger_script.setFocus(QtCore.Qt.OtherFocusReason)
         else:
             self.ui.tagger_script.setEnabled(False)
             self.ui.tagger_script.setText("")
@@ -201,13 +226,6 @@ class ScriptingOptionsPage(OptionsPage):
 
         self.restore_state()
 
-        args = {
-            "picard-doc-scripting-url": PICARD_URLS['doc_scripting'],
-        }
-        text = _('<a href="%(picard-doc-scripting-url)s">Open Scripting'
-                 ' Documentation in your browser</a>') % args
-        self.ui.scripting_doc_link.setText(text)
-
     def _all_scripts(self):
         for row in range(0, self.ui.script_list.count()):
             item = self.ui.script_list.item(row)
